# -*- coding: utf-8 -*-

"""
    Implementation of database interface for MLFP dataset.
    This protocol caters to only NIR subset of MLFP dataset.
"""

# Imports
import json
import os
import logging
import pkg_resources

from bob.pad.base.database import FileListPadDatabase
from bob.pad.face.database import VideoPadFile
from bob.bio.video import FrameSelector, FrameContainer
from bob.io.base import HDF5File

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

#------------------------------------------------------------------------------

class File(VideoPadFile):
    """Defines a File class for MLFP objects."""

    def __init__(self, attack_type, client_id, path, file_id=None):
        """Initialize the file object with details."""

        super(File, self).__init__(attack_type, client_id, path, file_id)

#------------------------------------------------------------------------------

    def load(self, directory=None, extension=None,\
            frame_selector=FrameSelector(selection_style='all')):
        """Load the file into FrameContainer."""

        path = self.make_path(directory=directory, extension=extension)

        with HDF5File(path) as f:
            return FrameContainer(hdf5=f)

#------------------------------------------------------------------------------


class MLFPDatabase(FileListPadDatabase):

    """A high level implementation of the Database class for NIR content of the
    MLFP dataset."""

    def __init__(
            self,
            name="MLFP NIR",
            original_directory=None,
            original_extension=".hdf5",
            annotation_directory=None,
            protocol="cv1",
            pad_file_class=File,
            **kwargs
            ):
        """Instantiate the MLFP database."""

        filelists_directory = pkg_resources.resource_filename(__name__, "/lists/mlfp/")
        self.filelists_directory = filelists_directory
        self.protocol = protocol

        super(MLFPDatabase, self).__init__(
            filelists_directory=filelists_directory,
            name=name,
            original_directory=original_directory,
            original_extension=original_extension,
            annotation_directory=annotation_directory,
            protocol=protocol,
            pad_file_class=pad_file_class,
            **kwargs
            )

        self.annotation_directory = annotation_directory

        logger.info("Database: {}".format(self.name))
        logger.info("Original directory: {}; Annotation directory: {}"\
            .format(self.original_directory, self.annotation_directory))

#------------------------------------------------------------------------------

    def annotations(self, f):

        """Return annotations for a given file object ``f``.
        Annotations must be precomputed."""

        if self.annotation_directory is None:
            raise ValueError("Annotation Directory is not provided.")

        file_path = os.path.join(self.annotation_directory, f.path+".json")

        if os.path.isfile(file_path):

            with open(file_path, "r") as json_file:
                annotations = json.load(json_file)

            if not annotations:  # if dictionary is empty
                logger.warning("Empty annotations for %s", f.path)
                return None

            return annotations

        else:
            logger.warning("Annotation file for %s does not exist.\
                (Overall path: %s)", f.path, file_path)

            return None

#------------------------------------------------------------------------------
