# -*- coding: utf-8 -*-

"""Implementation of dataset interface of WMCA Masks for PAD."""

# Imports
import os
import json
import logging
import pkg_resources

from bob.pad.base.database import FileListPadDatabase
from bob.pad.face.database.batl import BatlPadFile
from bob.db.batl.models import VideoFile

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

#------------------------------------------------------------------------------

class File(VideoFile):
    """File class for WMCA objects."""

    def __init__(self, path, client_id, session_id, presenter_id, type_id, pai_id):
        """Initialize the file objects with category and type details."""

        super(File, self).__init__(
            path=path,
            client_id=client_id,
            session_id=session_id,
            presenter_id=presenter_id,
            type_id=type_id,
            pai_id=pai_id
            )

        self.id = path

#------------------------------------------------------------------------------

class WMCAMask(FileListPadDatabase):

    """A high level implementation of the Database class for the WMCA Mask PAD
    database."""

    def __init__(
            self,
            name="WMCAMask",
            original_directory=None,
            original_extension=".h5",
            protocol="grandtest",
            annotation_directory=None,
            pad_file_class=BatlPadFile,
            low_level_pad_file_class=File,
            **kwargs):

        """
        Parameters
        -----------

        original_directory : str or None
            original directory refers to the location of WMCA (or WMCA Mask)
            parent directory.

        original_extension : str or None
            extension of original data.

        protocol : str (default: ``grandtest``)
            The protocol for which the clients should be retrieved.

        annotation_directory : str or None
            annotation directory refers to the location of precomputed
            annotations.

        """

        filelists_directory = pkg_resources.resource_filename(__name__, "lists/wmca_mask/")
        self.filelists_directory = filelists_directory

        # init the parent class using super.
        super(WMCAMask, self).__init__(
            filelists_directory=filelists_directory,
            name=name,
            protocol=protocol,
            original_directory=original_directory,
            original_extension=original_extension,
            pad_file_class=low_level_pad_file_class,
            annotation_directory=annotation_directory,
            **kwargs)

        self.low_level_pad_file_class = low_level_pad_file_class
        self.pad_file_class = pad_file_class
        self.annotation_directory = annotation_directory
        self.protocol = protocol

        logger.info("Dataset: {}".format(self.name))
        logger.info("Original directory: {}; Annotation directory: {}"\
            .format(self.original_directory, self.annotation_directory))

#------------------------------------------------------------------------------

    # override the _make_pad function in bob.pad.base

    def _make_pad(self, files):
        """Convert the file to PAD format."""

        video_pad_files = []

        for f in files:

            path = f.path
            client_id = f.client_id
            info = path.split("/")[-1]
            session_id = int(info.split("_")[1])
            presenter_id = int(info.split("_")[2])
            type_id = int(info.split("_")[3])
            pai_id = int(info.split("_")[4])

            video_file = self.low_level_pad_file_class(path=path,\
                client_id=client_id, session_id=session_id,\
                presenter_id=presenter_id, type_id=type_id,\
                pai_id=pai_id)

            video_pad_files.append(video_file)

        return video_pad_files

#------------------------------------------------------------------------------

    def objects(self, groups=None, protocol=None, purposes=None,\
            model_ids=None, **kwargs):
        """Retrieve file objects as per requirements."""

        # default the parameters if the values are not provided
        if protocol is None:
            protocol = self.protocol

        if groups is None:
            groups = ["train", "dev", "eval"]

        # obtain the file list using the parent class's functionality
        files = super(WMCAMask, self).objects(groups=groups, protocol=protocol,\
            purposes=purposes, model_ids=model_ids, **kwargs)

        # create objects for each file where the class is BATLPadFile

        # aux information
        stream_type = "infrared" # nir => infrared
        num_frames = 50

        files = [self.pad_file_class(f=f, stream_type=stream_type,\
            max_frames=num_frames) for f in files]

        return files

#------------------------------------------------------------------------------

    def annotations(self, f):
        """
        Return annotations for a given file object ``f``.
        Annotations must be precomputed."""

        if self.annotation_directory is None:
            raise ValueError("Annotation Directory is not provided.")

        file_path = os.path.join(self.annotation_directory, f.f.path+".json")

        if os.path.isfile(file_path):

            with open(file_path, "r") as json_file:
                annotations = json.load(json_file)

            if not annotations:  # if dictionary is empty
                logger.warning("Empty annotations for %s", f.path)
                return None

            return annotations

        else:
            logger.warning("Annotation file for %s does not exist", f.path)
            return None

#------------------------------------------------------------------------------
