# -*- coding: utf-8 -*-

""" Implementation of Patch Pooling CNN feature extractor with LightCNN-9
    backbone."""

# Imports
import torch
import logging
import numpy as np
from collections import OrderedDict

from PIL import Image
import torch.nn.functional as F
from torch.autograd import Variable
import torchvision.transforms as transforms

from bob.paper.nir_patch_pooling.extractor import LightCNN9
from bob.bio.base.extractor import Extractor

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

#------------------------------------------------------------------------------

class PatchPoolingCNN(Extractor):
    """Class to extract `patch pooled` features from the final convolutional
        layer of LightCNN9 (MFM5 layer)."""

    def __init__(self, model_file=None, patch_stride=4, num_classes=79077):
        """Initialize with default parameters."""

        Extractor.__init__(self, skip_extractor_training=True)
        self.network = LightCNN9Patch()
        self.patch_stride = patch_stride

        # load the model into network.
        cp = torch.load(model_file, map_location="cpu")
        cp = cp["state_dict"]

        # checked if pre-trained model was saved using nn.DataParallel
        saved_with_nn_parallel = any("module" in x for x in list(cp.keys()))

        # if DataParallel format, remove module term
        if(saved_with_nn_parallel):
            new_state_dict = OrderedDict()

            for k, v in cp.items():
                k_new = k[7:]
                new_state_dict[k_new] = v

            self.network.load_state_dict(new_state_dict)

        else:

            self.network.load_state_dict(cp)

        self.network.eval()

        # image pre-processing
        self.data_transform = transforms.Compose([transforms.Resize(size=128),\
            transforms.ToTensor()])

#------------------------------------------------------------------------------

    def __call__(self, image):
        """ Extract features from an image."""

        # torchvision.transforms expect a numpy array of size HxWxC
        pil_image = Image.fromarray(image.astype(np.uint8))
        input_image = self.data_transform(pil_image)

        input_image = input_image.unsqueeze(0).float()
        input_image = Variable(input_image)

        # obtain the features (to be pooled) from forward pass of network
        _, features = self.network(input_image)

        # pool features through patch-level processing
        features = self.conv_to_patch(features)
        features = features.data.numpy().flatten()

        return features.astype(np.float64)

#------------------------------------------------------------------------------

    def conv_to_patch(self, features):
        """Pooling of features over local patches."""

        stride = self.patch_stride
        pooled_features = torch.zeros(1, stride*stride*features.shape[1])

        # obtain patches by tesselation of feature maps
        # pool linearized version of individual patches
        for i in range(0, features.shape[2], stride):
            for j in range(0, features.shape[3], stride):

                feat_temp = features[:, :, i:i+stride, j:j+stride]
                feat_temp = feat_temp.contiguous().view(feat_temp.size(0), -1)
                pooled_features += feat_temp

        # normalize the vector of pooled features
        pooled_features = pooled_features/stride/stride

        return pooled_features

#------------------------------------------------------------------------------


class LightCNN9Patch(LightCNN9):
    """Implementation of Patch-formatted LightCNN0."""

    def __init__(self):

        # do not change the init
        super(LightCNN9Patch, self).__init__()

#------------------------------------------------------------------------------

    def forward(self, x):

        """Propagate data through the network, and return the output of final
        conv layer."""

        conv_out = self.features(x)
        x = conv_out.view(conv_out.size(0), -1)
        x = self.fc1(x)
        x = F.dropout(x, training=self.training)
        out = self.fc2(x)
        return out, conv_out

#------------------------------------------------------------------------------
