"""
   Desc: Script to generate face annotations for NIR data
   Author: Ketan Kotwal
"""

#------------------------------------------------------------------------------

# imports
import json
import sys
import os
import numpy as np
from bob.pad.face.preprocessor.FaceCropAlign import detect_face_landmarks_in_image
from bob.io.base import create_directories_safe
from bob.ip.color import rgb_to_gray
from bob.ip.facelandmarks import detect_landmarks
from bob.paper.nir_patch_pooling.config.wmca_mask import database as d_wmca
from bob.paper.nir_patch_pooling.config.mlfp import database as d_mlfp

#------------------------------------------------------------------------------

def bob_annotate(image):
    """Find face landmarks using bob's routine."""

    image = rgb_to_gray(image)
    lm_list = detect_landmarks(image, 1)
    lm = lm_list[0].landmarks
    bounding_box = lm_list[0].bounding_box
    annotations = {}

    if lm is not None:
        lm = np.array(lm)
        lm = np.vstack((lm[:, 1], lm[:, 0])).T
        right_eye, left_eye = _get_eye_pos(lm)

        points = []
        for i in range(lm.shape[0]):
            points.append((int(lm[i, 0]), int(lm[i, 1])))

        annotations['topleft'] = bounding_box.topleft
        annotations['bottomright'] = bounding_box.bottomright
        annotations['landmarks'] = points
        annotations['leye'] = left_eye
        annotations['reye'] = right_eye

    return annotations

#------------------------------------------------------------------------------

def _get_eye_pos(lm):
    """Return the locations of left and right eyes."""

    rr_c = (lm[36, :]+lm[39, :])/2.0
    ll_c = (lm[42, :]+lm[45, :])/2.0

    right_eye = (int(rr_c[1]), int(rr_c[0]))
    left_eye = (int(ll_c[1]), int(ll_c[0]))

    return right_eye, left_eye

#------------------------------------------------------------------------------

def find_annotations(image):
    """Find face landmarks for image using variety of methods."""

    # if image is grayscale, convert to 3 channel for face detection
    if len(image.shape) == 2:
        image = np.repeat(image[:, :, np.newaxis], 3, axis=2)
        image = np.transpose(image, (2, 0, 1))

    # find annotations using MTCNN
    frame_annotations = detect_face_landmarks_in_image(image, method="mtcnn")
    if frame_annotations:
        return frame_annotations

    # else, find annotations using dlib
    frame_annotations = detect_face_landmarks_in_image(image, method="dlib")
    if frame_annotations:
        return frame_annotations

    # else, find annotations using landmark detection routines from bob
    #frame_annotations = bob_annotate(image.astype(np.uint8))
    #if frame_annotations:
    #    return frame_annotations

    # else, return empty dictionary with warning
    print(" --> Could not find annotations")
    return None

#------------------------------------------------------------------------------

def normalize_image(image, n_sigma=3.0, use_central_region=False):
    """Normalize image using MAD."""

    assert(len(image.shape) == 2)
    image = image.astype(np.float64)

    # use central region of image to determine parameters for normalization
    if use_central_region:
        h, w = image.shape
        region = image[int(0.25*h):int(0.75*h), int(0.25*w):int(0.75*w)]
    else:
        region = image

    # calculate median values
    med = np.median(region)
    mad = np.median(np.abs(region - med))

    image_n = ((image-med+n_sigma*mad)/(2.0*n_sigma*mad))*255.0

    # Clamping to 0-255
    image_n = np.maximum(image_n, 0)
    image_n = np.minimum(image_n, 255)

    image_n = image_n.astype(np.uint8)

    return image_n

#------------------------------------------------------------------------------

class Annotator(object):
    """A class to generate annotations for bob databases."""

    def __init__(self, database, annotation_directory, n_sigma=3.0,\
            use_central_region=False):
        """Instantiate the Annotator class with database object (compatible to
        bob), annotation directory, and parameters for image normalization."""

        self.database = database
        self.annotation_directory = annotation_directory
        self.n_sigma = n_sigma
        self.use_central_region = use_central_region

        if not os.path.exists(self.annotation_directory):
            os.makedirs(self.annotation_directory)
            print("Annotation directory created at:{}"\
                .format(self.annotation_directory))

#------------------------------------------------------------------------------

    def process_frame(self, fc):
        """Process the FrameContainer for annotations."""

        annotations = {}
        prev_valid_index = None

        for index, frame, _ in fc:

            image = normalize_image(frame, n_sigma=self.n_sigma,\
                use_central_region=self.use_central_region)

            try:
                frame_annotations = find_annotations(image)

            except Exception as e:
                print("Exception: {}".format(e))

                # copy annotations of previous frame
                if (prev_valid_index is not None):
                    frame_annotations = annotations[prev_valid_index]
                    print(" --> Copying annotations of valid previous frame")

                else:
                    frame_annotations = None
                    print(" --> Setting empty annotations")

            if frame_annotations is not None:
                annotations[str(index)] = frame_annotations
                prev_valid_index = str(index)

        return annotations

#------------------------------------------------------------------------------

    def process_video(self, f):
        """Processe a PAD video for annotations."""

        # load the video into framecontainer
        fc = f.load(directory=self.database.original_directory, extension=\
            self.database.original_extension)

        # obtain the annotations
        annotations = self.process_frame(fc)

        # save the annotations as json.
        json_filepath = os.path.join(self.annotation_directory, f.path + ".json")

        create_directories_safe(directory=os.path.split(json_filepath)[0], dryrun=False)
        with open(json_filepath, "w+") as json_file:
            json_file.write(json.dumps(annotations))

        return

#------------------------------------------------------------------------------

    def run(self, job_index):
        """Execute the annotation script over the entire dataset."""

        # collect the files to be processed
        self.filelist = self.database.objects(protocol=self.database.protocol)
        total = len(self.filelist)
        print("Files to be annotated: {}".format(total))

        # calculate split indices if computation is parallel
        if(job_index != -1):
            num_jobs = 32
            file_range = np.linspace(0, total, num_jobs+1, dtype=np.int)
            start_index = file_range[job_index]
            end_index = file_range[job_index+1]
        else:
            start_index = 0
            end_index = total

        print("Processing Job Index: {} (Files: {} to {})".format(job_index,\
            start_index, end_index))

        # process each video in the given range
        for idx, f in enumerate(self.filelist[start_index:end_index]):

            print("[{:03d}/{:03d}] Sample: {}".format(idx+1, total, f.path))
            json_path = os.path.join(self.annotation_directory, f.path +  ".json")

            if not os.path.exists(json_path):
                self.process_video(f)

            else:
                print("Annotations exist: {}. Skipping".format(f.path))

        return

#------------------------------------------------------------------------------

def main():
    """A script to annotate the bob dataset.

    This scripts generates the annotations for frames in each video, stores
    the annotations in json file (one per video) in the annotation_directory.
    The number of frames selected per video can be set through the dataset
    interface.

    Parameters
    ----------
    database_name : str
        The name of database. (``wmca`` or ``mlfp``)
    annotation_directory : str
        The desired location of annotation directory.
    n_sigma : float
        The sigma value for MAD normalization (default: 3.0)
    use_central_region : bool
        Set to True if face should be searched only within the central half
        region of the frame.

    Returns
    -------
    None

    """

    if len(sys.argv) < 3:
        print("Usage: {} <database_name> <annotation-directory> [<job-index>]"\
            .format(__name__))
        exit(0)

    if sys.argv[1] == "wmca":
        database = d_wmca
    elif sys.argv[1] == "mlfp":
        database = d_mlfp
        #database.protocol = "cv1"
    else:
        raise ValueError("database must be either wmca or mlfp.")

    annotation_directory = sys.argv[2]
    print("Database: {}. Annotation directory: {}".format(database.name,\
        annotation_directory))

    if len(sys.argv) >= 4:
        job_index = int(float(sys.argv[3]))
        print("# Job Index: {}".format(job_index))
    else:
        job_index = -1

    if sys.argv[1] == "wmca":
        an = Annotator(database, annotation_directory)
    else: # for MLFP
        an = Annotator(database, annotation_directory, n_sigma=4.0,\
        use_central_region=True)

    an.run(job_index)

#------------------------------------------------------------------------------

if __name__ == "__main__":

    main()

#------------------------------------------------------------------------------
