"""
    Desc: Script to convert MLFP (NIR) data into specific format compatible
    to bob
    Author: Ketan Kotwal
"""

#------------------------------------------------------------------------------

# imports
import os
import sys
import numpy as np
import scipy.io as spio
from bob.bio.video import FrameContainer
from bob.io.base import create_directories_safe, HDF5File

FRAMES_PER_VIDEO = 20

#------------------------------------------------------------------------------

class MLFPConvertor(object):
    """A class to convert the NIR files MLFP dataset into bob compatible format."""

    def __init__(self, input_directory, output_directory):
        """Instantiate the class with input and output directories."""

        self.input_directory = input_directory

        self.output_directory = output_directory
        if not os.path.exists(self.output_directory):
            os.makedirs(self.output_directory)

        self.file_objects = self._load_db(self.input_directory)

        print("Input directory: {}".format(self.input_directory))
        print("Output directory: {}".format(self.output_directory))

#------------------------------------------------------------------------------

    def _load_db(self, db_directory):
        """Create a list of files by walkthrough into input directory."""

        file_list = []

        for dirpath, _, files in os.walk(db_directory):
            for name in files:

                file_path = os.path.join(dirpath, name)
                file_path = file_path[43:]
                if ("Infrared.mat" in file_path):
                    file_list.append(file_path[:-4])

        return file_list

#------------------------------------------------------------------------------

    def process_file(self, filename):
        """Transform a .mat file into .hdf5 file containing FrameContainer."""

        # load the file
        fc = spio.loadmat(os.path.join(self.input_directory, filename + ".mat"))
        fc = fc["IR"]

        # select first frames_per_video frames and add to framecontainer
        fc_out = FrameContainer()

        for idx, frame in enumerate(fc[:FRAMES_PER_VIDEO]):
            frame = frame[0]/256.0
            fc_out.add(idx, frame.astype(np.uint8), None)

        # save fc to hdf file
        out_filepath = os.path.join(self.output_directory, filename + ".hdf5")
        create_directories_safe(directory=os.path.split(out_filepath)[0], dryrun=False)

        hdf_out = HDF5File(out_filepath, 'w')
        fc_out.save(hdf_out)
        del hdf_out

        return

#------------------------------------------------------------------------------

    def run(self):
        """Execute the conversion script."""

        total = len(self.file_objects)
        print("Files to be converted to .hdf5: {}".format(total))

        for idx, f in enumerate(self.file_objects):

            print("[{:03d}/{:03d}] Sample: {}".format(idx+1, total, f))

            out_filepath = os.path.join(self.output_directory, f + ".hdf5")
            if not os.path.exists(out_filepath):
                self.process_file(f)
            else:
                print("File exist for {}. Skipping".format(f))

        return

#------------------------------------------------------------------------------

def main():
    """A script to convert the MLFP dataset into bob compatible format.

    This script converts the MLFP dataset (only NIR files) into bob compatible
    dataset format. The script can function properly only if the dataset
    structure is unchanged.

    Parameters
    ----------
    input_directory : str
        The location of the MLFP dataset
    output_directory : str
        The desired location of converted data

    Returns
    -------
    None

    """

    input_directory = sys.argv[1]
    output_directory = sys.argv[2]

    m_conv = MLFPConvertor(input_directory, output_directory)
    m_conv.run()

#------------------------------------------------------------------------------

if __name__ == "__main__":

    main()

#------------------------------------------------------------------------------
